/*
 * Copyright 2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Header:$
 */
package org.apache.beehive.netui.pageflow;

import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.beehive.netui.util.internal.InternalStringBuilder;
import org.apache.beehive.netui.util.logging.Logger;

import javax.servlet.ServletContext;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

/**
 * Default event reporter.  Logs every event when the log level is set to "debug" or "trace".
 */ 
public class DefaultPageFlowEventReporter
    extends PageFlowEventReporter
{
    private static final Logger _log = Logger.getInstance( DefaultPageFlowEventReporter.class );
    
    protected DefaultPageFlowEventReporter( ServletContext servletContext )
    {
        super( servletContext );
    }

    public void actionRaised( FlowController flowController, ActionMapping mapping, ActionForm form,
                              ServletRequest request, ServletResponse response )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "actionRaised" );
            msg.addParam( "FlowController", flowController );
            msg.addParam( "ActionMapping", mapping );
            msg.addParam( "ActionForm", form );
            msg.addParam( "Request", request );
            msg.addParam( "Response", response );
            _log.debug( msg );
        }
    }

    public void actionSuccess( FlowController flowController, ActionMapping mapping, ActionForm form,
                               ServletRequest request, ServletResponse response, ActionForward result,
                               long timeTakenMillis )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "actionSuccess" );
            msg.addParam( "FlowController", flowController );
            msg.addParam( "ActionMapping", mapping );
            msg.addParam( "ActionForm", form );
            msg.addParam( "Request", request );
            msg.addParam( "Response", response );
            msg.addParam( "ActionForward", result );
            msg.addParam( "TimeTakenMillis", new Long( timeTakenMillis ) );
            _log.debug( msg );
        }
    }

    public void exceptionRaised( Throwable ex, ActionMapping actionMapping, ActionForm form,
                                 FlowController flowController, ServletRequest request,
                                 ServletResponse response )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "exceptionRaised" );
            msg.addParam( "Throwable", ex );
            msg.addParam( "ActionMapping", actionMapping );
            msg.addParam( "ActionForm", form );
            msg.addParam( "FlowController", flowController );
            msg.addParam( "Request", request );
            msg.addParam( "Response", response );
            _log.debug( msg );
        }
    }

    public void exceptionHandled( Throwable ex, ActionMapping actionMapping, ActionForm form,
                                  FlowController flowController, ServletRequest request,
                                  ServletResponse response, ActionForward result,
                                  long timeTakenMillis )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "exceptionHandled" );
            msg.addParam( "Throwable", ex );
            msg.addParam( "ActionMapping", actionMapping );
            msg.addParam( "ActionForm", form );
            msg.addParam( "FlowController", flowController );
            msg.addParam( "Request", request );
            msg.addParam( "Response", response );
            msg.addParam( "ActionForward", result );
            msg.addParam( "TimeTakenMillis", new Long( timeTakenMillis ) );
            _log.debug( msg );
        }
    }

    public void flowControllerCreated( FlowController flowController, ServletRequest request,
                                       ServletResponse response )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "flowControllerCreated" );
            msg.addParam( "FlowController", flowController );
            msg.addParam( "Request", request );
            msg.addParam( "Response", response );
            _log.debug( msg );
        }
    }

    public void flowControllerDestroyed( FlowController flowController )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "flowControllerDestroyed" );
            msg.addParam( "FlowController", flowController );
            _log.debug( msg );
        }
    }

    public void beginActionRequest( ServletRequest request, ServletResponse response )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "beginActionRequest" );
            msg.addParam( "Request", request );
            msg.addParam( "Response", response );
            _log.debug( msg );
        }
    }

    public void endActionRequest( ServletRequest request, ServletResponse response, long timeTakenMillis )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "endActionRequest" );
            msg.addParam( "Request", request );
            msg.addParam( "Response", response );
            msg.addParam( "TimeTakenMillis", new Long( timeTakenMillis ) );
            _log.debug( msg );
        }
    }

    public void beginPageRequest( ServletRequest request, ServletResponse response )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "beginPageRequest" );
            msg.addParam( "Request", request );
            msg.addParam( "Response", response );
            _log.debug( msg );
        }
    }

    public void endPageRequest( ServletRequest request, ServletResponse response, long timeTakenMillis )
    {
        if ( _log.isDebugEnabled() )
        {
            LogMsg msg = new LogMsg( "endPageRequest" );
            msg.addParam( "Request", request );
            msg.addParam( "Response", response );
            msg.addParam( "TimeTakenMillis", new Long( timeTakenMillis ) );
            _log.debug( msg );
        }
    }
        
    protected static class LogMsg
    {
        private String _eventName;
        private InternalStringBuilder _logMessage;
            
        public LogMsg( String eventName )
        {
            _eventName = eventName;
        }
            
        public void addParam( String name, Object value )
        {
            if ( _logMessage == null )
            {
                _logMessage = new InternalStringBuilder( _eventName ).append( ": " );
            }
            else
            {
                _logMessage.append( ", " );
            }
                
            _logMessage.append( name ).append( '=' ).append( value );
        }
            
        public String toString()
        {
            return _logMessage == null ? _eventName : _logMessage.toString();
        }
    }
}
