/*
 * Copyright 2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Header:$
 */
package org.apache.beehive.netui.pageflow;

import org.apache.beehive.netui.core.urltemplates.URLTemplateDescriptor;
import org.apache.beehive.netui.pageflow.handler.Handlers;
import org.apache.beehive.netui.pageflow.internal.AdapterManager;
import org.apache.beehive.netui.pageflow.internal.InternalConstants;
import org.apache.beehive.netui.pageflow.internal.LegacySettings;
import org.apache.beehive.netui.util.config.ConfigInitializationException;
import org.apache.beehive.netui.util.config.ConfigUtil;
import org.apache.beehive.netui.util.config.bean.PrefixHandlers;
import org.apache.beehive.netui.util.logging.Logger;

import javax.servlet.ServletContext;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;
import java.io.IOException;
import java.io.InputStream;



/**
 * Performs various initialization at ServletContext-init time.
 */ 
public class PageFlowContextListener
        implements ServletContextListener
{
    private static final String ALREADY_INIT_ATTR = InternalConstants.ATTR_PREFIX + "contextInit";
            
    private static final Logger _log = Logger.getInstance( PageFlowContextListener.class );
    
    
    public void contextInitialized( ServletContextEvent event )
    {
        performInitializations( event.getServletContext() );
    }

    public void contextDestroyed( ServletContextEvent event )
    {
    }
    
    static boolean isInit( ServletContext servletContext )
    {
        return servletContext.getAttribute( ALREADY_INIT_ATTR ) != null;
    }
    
    static void performInitializations( ServletContext servletContext )
    {
        servletContext.setAttribute( ALREADY_INIT_ATTR, Boolean.TRUE );
        
        //
        // Initialize the config file, unless it's already initialized.  This can happen because the scope for the 
        // config (static) isn't the same as the scope for PageFlowActionServlet, which may get created and destroyed
        // within a classloader (which is the case during StrutsTestCase tests).
        //
        if ( ! ConfigUtil.isInit() )
        {
            try
            {
                InputStream configInput = servletContext.getResourceAsStream( InternalConstants.NETUI_CONFIG_PATH );
                
                try
                {
                    ConfigUtil.init( configInput );
                }
                finally
                {
                    try
                    {
                        if ( configInput != null ) configInput.close();
                    }
                    catch ( IOException e )
                    {
                        if ( _log.isErrorEnabled() )
                        {
                            _log.error( "Could not close input for " + InternalConstants.NETUI_CONFIG_PATH );
                        }
                    }
                }
            }
            catch ( ConfigInitializationException e )
            {
                _log.fatal( "Could not initialize from " + InternalConstants.NETUI_CONFIG_PATH, e );
                throw new IllegalStateException( "Could not initialize from " + InternalConstants.NETUI_CONFIG_PATH, e );
            }
        }
        
        AdapterManager.initServletContext( servletContext );
        LegacySettings.init( servletContext );
        Handlers.init( servletContext );
        URLTemplateDescriptor.getInstance().load( servletContext );
        initPrefixHandlers();
    }

    /**
     * This method will initialize all of the PrefixHandlers registered in the netui config.
     * The prefix handlers are registered with ProcessPopulate and are typically implemented as
     * public inner classes in the tags that require prefix handlers.
     */
    private static void initPrefixHandlers()
    {
        PrefixHandlers ph = ConfigUtil.getConfig().getPrefixHandlers();
        if ( ph == null )
        {
            return;
        }
        PrefixHandlers.PrefixHandler[] prefixHandlers = ph.getPrefixHandlerArray();
        if ( prefixHandlers != null )
        {
            for ( int i = 0; i < prefixHandlers.length; i++ )
            {
                try
                {
                    Class prefixClass = Class.forName( prefixHandlers[i].getHandlerClass() );
                    String name = prefixHandlers[i].getName();
                    if ( name == null || name.equals( "" ) )
                    {
                        _log.warn( "The name for the prefix handler '" + prefixHandlers[i].getHandlerClass()
                                   + "' must not be null" );
                        continue;
                    }
                    Object o = prefixClass.newInstance();
                    if ( !( o instanceof RequestParameterHandler ) )
                    {
                        _log.warn( "The class '" + prefixHandlers[i].getHandlerClass()
                                   + "' must be an instance of RequestParameterHandler" );
                        continue;
                    }
                    ProcessPopulate.registerPrefixHandler( name, ( RequestParameterHandler ) o );
                }
                catch ( ClassNotFoundException e )
                {
                    _log.warn( "Class '" + prefixHandlers[i].getHandlerClass() + "' not found", e );
                }
                catch ( IllegalAccessException e )
                {
                    _log.warn( "Illegal access on Class '" + prefixHandlers[i].getHandlerClass() + "'", e );

                }
                catch ( InstantiationException e )
                {
                    _log.warn( "InstantiationException on Class '" + prefixHandlers[i].getHandlerClass() + "'",
                               e.getCause() );
                }
            }
        }
    }
}
