/*
 * Copyright 2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Header:$
 */
package org.apache.beehive.netui.core.urls;

import org.apache.beehive.netui.core.urltemplates.URLTemplate;
import org.apache.beehive.netui.core.urltemplates.URLTemplateDescriptor;
import org.apache.beehive.netui.util.config.bean.UrlConfig;
import org.apache.beehive.netui.util.config.ConfigUtil;

import javax.servlet.ServletRequest;


/**
 * Default implementation of TemplatedURLFormatter for formatting URLs
 * based on templates from a URL template config file.
 */
public class DefaultTemplatedURLFormatter implements TemplatedURLFormatter
{
    private static final String DEFAULT_TEMPLATE_REF = "default-url-templates";

    /**
     * Format the given URL using a URL template, if defined in a URL
     * template config file. The {@link URIContext}
     * encapsulates some additional data needed to write out the string form.
     * E.g. It defines if the &quot;&amp;amp;&quot; entity or the
     * '&amp;' character should be used to separate quary parameters.
     *
     * @param request the current ServletRequest.
     * @param uri the MutableURI to be formatted into a String.
     * @param key key for the URL template type to use for formatting the URI
     * @param uriContext data required to write out the string form.
     * @return the URL as a <code>String</code>
     */
    public String getTemplatedURL( ServletRequest request, MutableURI uri, String key, URIContext uriContext )
    {
        // Look for the template config and get the right template.
        // If it is found, apply the value to the template.
        String result = null;
        String templateName = URLTemplateDescriptor.getInstance().getURLTemplateRef( DEFAULT_TEMPLATE_REF, key );

        if ( templateName != null )
        {
            URLTemplate template = URLTemplateDescriptor.getInstance().getURLTemplate( templateName );
            result = formatURIWithTemplate( request, uri, uriContext, template );
        }
        else
        {
            // no template found, just return the uri as a String...
            result = uri.getURIString( uriContext );
        }

        return result;
    }

    private String formatURIWithTemplate( ServletRequest request, MutableURI uri,
                                          URIContext uriContext, URLTemplate template )
    {
        String scheme = uri.getScheme();
        String host = uri.getHost();
        int port = uri.getPort();

        if ( scheme == null || scheme.length() == 0 ) { scheme = request.getScheme(); }

        if ( host == null || host.length() == 0 ) { host = request.getServerName(); }

        if ( port < 0 ) { port = request.getServerPort(); }

        template.substitute( URLTemplateDescriptor.SCHEME_TOKEN, scheme );
        template.substitute( URLTemplateDescriptor.DOMAIN_TOKEN, host );
        template.substitute( URLTemplateDescriptor.PORT_TOKEN, port );
        template.substitute( URLTemplateDescriptor.PATH_TOKEN, uri.getPath() );

        String query = null;
        query = uri.getQuery( uriContext );

        if ( query == null ) {
            query = "";
        }

        template.substitute( URLTemplateDescriptor.QUERY_STRING_TOKEN, query );

        String fragment = uri.getFragment();

        if ( fragment == null ) { fragment = ""; }

        template.substitute( URLTemplateDescriptor.FRAGMENT_TOKEN, fragment );

        return template.toString();
    }
}
