/*
 * Copyright 2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Header:$
 */
package org.apache.beehive.netui.core.urltemplates;

import org.apache.beehive.netui.util.internal.InternalStringBuilder;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;


/**
 * The class to format a URL defined by url-template-config template
 * given by values for a set of tokens.
 */
public class URLTemplate
{
    private static final char BEGIN_TOKEN_QUALIFIER = '{';
    private static final char END_TOKEN_QUALIFIER = '}';

    // The String form of the template.
    private String _template;

    // Parsed representation of the template... list of literals and tokens.
    private ArrayList/*< TemplateItem >*/ _parsedTemplate = null;

    private boolean _isParsed = false;

    // The set of known tokens in a valid template.
    private Collection/*< String >*/ _knownTokens = null;

    // The set of required tokens in a valid template.
    private Collection/*< String >*/ _requiredTokens = null;

    private HashMap/*< String, String >*/ _tokenValuesMap = new HashMap/*< String, String >*/();

    /**
     * Create a URLTemplate from a url-template-config template.
     *
     * <p> Allow clients to define a set of required and known tokens for the
     * template verification. Tokens are expected to be qualified
     * in braces. E.g. {url:path} </p>
     *
     * <p> The template verification will ensure the URL template conforms to
     * a valid format for known tokens and contains the required tokens. </p>
     *
     * <p> Should call verify after creating a new template. </p>
     *
     * @param template the string form of the template from url-template-config.
     * @param knownTokens The set of known tokens for a valid template.
     * @param requiredTokens The set of required tokens in a valid template.
     */
    public URLTemplate( String template, Collection/*< String >*/ knownTokens, Collection/*< String >*/ requiredTokens )
    {
        setTemplate( template );
        _knownTokens = knownTokens;
        _requiredTokens = requiredTokens;
    }

    /**
     * Copy constructor to create a URLTemplate from an existing URLTemplate.
     *
     * <p> Note that this is not truly a complete copy because the Map
     * of the replacement values for the given tokens is not copied.
     * This copy will just have an empty map of token values so that
     * it is "cleared" and ready to format another URL. </p>
     *
     * @param template the URLTemplate to copy.
     */
    public URLTemplate( URLTemplate template )
    {
        setTemplate( template.getTemplate() );
        _knownTokens = template._knownTokens;
        _requiredTokens = template._requiredTokens;
        _parsedTemplate = template._parsedTemplate;
        _isParsed = template._isParsed;
    }

    /**
     * Reset the String form of the template.
     *
     * <p> Should call verify after setting a new template. </p>
     *
     * @param template the string form of the template from url-template-config.
     */
    public void setTemplate( String template )
    {
        if ( template == null || template.length() == 0 )
        {
            throw new IllegalStateException( "Template cannot be null or empty." );
        }

        if ( template.equals( _template ) ) return;

        _template = template;
        _isParsed = false;
        _parsedTemplate = null;
    }

    /**
     * Retrieve the String form of the template.
     *
     * @return the string form of the template.
     */
    public String getTemplate()
    {
        return _template;
    }

    /**
     * Verification will ensure the URL template conforms to a valid format
     * for known tokens and contains the required tokens. It will also parse
     * the tokens and literal data into a list to improve the replacement
     * performance when constructing the final URL string.
     */
    public void verify() throws IllegalStateException
    {
        // For each known token, make sure there is a leading and trailing brace
        if ( _knownTokens != null )
        {
            for ( java.util.Iterator ii = _knownTokens.iterator(); ii.hasNext(); )  
            {
                String token = ( String ) ii.next();
                if ( token != null && token.length() > 2 )
                {
                    // Strip braces from the known token
                    token = token.substring( 1, token.length() - 1 );
                    int index = _template.indexOf( token );
                    if ( index != -1 )
                    {
                        if ( _template.charAt( index - 1 ) != BEGIN_TOKEN_QUALIFIER
                                || _template.charAt( index + token.length() ) != END_TOKEN_QUALIFIER )
                        {
                            throw new IllegalStateException( "Template token, " + token
                                    + ", is not correctly enclosed with braces in template: " + _template );
                        }
                    }
                }
            }
        }

        // Parse the template into tokens and literals
        parseTemplate();

        // Check if the required tokens are present
        if ( _requiredTokens != null )
        {
            for ( java.util.Iterator ii = _requiredTokens.iterator(); ii.hasNext(); )  
            {
                String token = ( String ) ii.next();
                String qualifiedToken = token;
                TemplateItem requiredItem = new TemplateItem( qualifiedToken, true );

                if ( !_parsedTemplate.contains( requiredItem ) )
                {
                    throw new IllegalStateException( "Required token, " + token
                            + ", not found in template: " + _template );
                }
            }
        }
    }

    private void parseTemplate()
    {
        if ( _isParsed ) return;

        _parsedTemplate = new ArrayList/*< TemplateItem >*/();
        TemplateTokenizer tokenizer = new TemplateTokenizer( getTemplate() );
        for ( ; tokenizer.hasNext(); )
        {
            boolean isToken = tokenizer.isTokenNext();
            String tokenOrLiteral = ( String ) tokenizer.next();
            TemplateItem item = new TemplateItem( tokenOrLiteral, isToken );
            _parsedTemplate.add( item );
        }

        _isParsed = true;
    }

    /**
     * Replace a set of tokens in the template with a corresponding set of values.
     * This assumes that there is an ordered one-to-one relationship. Tokens are
     * expected to be qualified in braces. E.g. {url:path}
     */
    public void substitute( Map/*< String, String >*/ tokensAndValues )
    {
        if ( tokensAndValues != null )
        {
            _tokenValuesMap.putAll( tokensAndValues );
        }
    }

    /**
     * Replace a single token in the template with a corresponding String value.
     * Tokens are expected to be qualified in braces. E.g. {url:path}
     */
    public void substitute( String token, String value )
    {
        _tokenValuesMap.put( token, value );
    }

    /**
     * Replace a single token in the template with a corresponding int value.
     * Tokens are expected to be qualified in braces. E.g. {url:port}
     */
    public void substitute( String token, int value )
    {
        String valueStr = Integer.toString( value );
        _tokenValuesMap.put( token, valueStr );
    }

    /**
     * Return the String representation of the template after
     * replacing all tokens with their associated values..
     */
    public String toString()
    {
        // template should already have been parsed with a call to
        if ( !_isParsed )
        {
            // Parse the template into tokens and literals
            parseTemplate();
        }

        InternalStringBuilder result = new InternalStringBuilder( _template.length() + 16 );
        for ( java.util.Iterator ii = _parsedTemplate.iterator(); ii.hasNext(); )  
        {
            TemplateItem item = ( TemplateItem ) ii.next();
            if ( item.isToken() )
            {
                if ( _tokenValuesMap.containsKey( item.getValue() ) )
                {
                    appendToResult( result, ( String ) _tokenValuesMap.get( item.getValue() ) );
                }
                else
                {
                    // No value for the token so treat the token as a literal.
                    appendToResult( result, item.getValue() );
                }
            }
            else
            {
                appendToResult( result, item.getValue() );
            }
        }

        return result.toString();
    }

    // check to make sure we don't end up with "//" between components
    // of the URL
    private void appendToResult( InternalStringBuilder result, String value )
    {
        if ( value == null || value.length() == 0 ) return;

        if ( result.length() > 0 && result.charAt( result.length() - 1 ) == '/'
                && value.charAt( 0 ) == '/' )
        {
            result.deleteCharAt( result.length() - 1 );
        }
        result.append( value );
    }

    protected class TemplateItem
    {
        private String value;
        private boolean isToken = false;

        public TemplateItem( String value, boolean isToken )
        {
            assert value != null : "TemplateItem value cannot be null.";
            this.value = value;
            this.isToken = isToken;
        }

        public String getValue()
        {
            return value;
        }

        public boolean isToken()
        {
            return isToken;
        }

        public boolean equals( Object o )
        {
            if ( this == o )
            {
                return true;
            }
            if ( !( o instanceof TemplateItem ) )
            {
                return false;
            }

            final TemplateItem templateItem = ( TemplateItem ) o;

            if ( isToken != templateItem.isToken() )
            {
                return false;
            }
            if ( !value.equals( templateItem.getValue() ) )
            {
                return false;
            }

            return true;
        }

        public int hashCode()
        {
            int result;
            result = value.hashCode();
            result = 29 * result + ( isToken ? 1 : 0 );
            return result;
        }
    }
}
