/*
 * Copyright 2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Header:$
 */
package org.apache.beehive.netui.core.urltemplates;

import org.apache.beehive.netui.core.urltemplates.schema.UrlTemplateConfigDocument.UrlTemplateConfig;
import org.apache.beehive.netui.core.urltemplates.schema.UrlTemplateDocument;
import org.apache.beehive.netui.core.urltemplates.schema.UrlTemplateRefGroupDocument;
import org.apache.beehive.netui.core.urltemplates.schema.UrlTemplateRefDocument;
import org.apache.beehive.netui.util.logging.Logger;
import org.apache.xmlbeans.XmlCursor;
import org.apache.xmlbeans.XmlException;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import java.io.InputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;

/**
 * Methods for configuring and retrieving the URLTemplates object.
 */
public class URLTemplatesFactory
{
    private static final Logger _log = Logger.getInstance( DefaultURLTemplateFileParser.class );

    /** Default value for path from the web app to the URL templates. */
    public static final String DEFAULT_URL_TEMPLATE_CONFIG_FILE_PATH = "/WEB-INF/beehive-url-template-config.xml";

    // Path to the URL templates.
    private String _configFilePath = DEFAULT_URL_TEMPLATE_CONFIG_FILE_PATH;

    // Parser to use on the URL template config file.
    private URLTemplateFileParser _parser = new DefaultURLTemplateFileParser();

    // The set of known tokens in a valid template.
    private Collection _knownTokens = null;

    // The set of required tokens in a valid template.
    private Collection _requiredTokens = null;

    /**
     * Allow clients to set their own URL template config file name or path.
     *
     * @param configFilePath An absolute path from the web app root tothe URL template config file.
     */
    public void setConfigFilePath( String configFilePath )
    {
        if ( configFilePath == null )
        {
            throw new IllegalArgumentException( "Config file path cannot be null." );
        }

        _configFilePath = configFilePath;
    }

    /**
     * Allow clients to set their own parser.
     * <p/>
     * The parser is used on the URL template config file,
     * WEB-INF/url-template-config.xml. This can be used to support older
     * file formats.
     *
     * @param parser The parser to use on the URL template config file.
     */
    public void setParser( URLTemplateFileParser parser )
    {
        if ( parser == null )
        {
            throw new IllegalArgumentException( "Parser cannot be null." );
        }

        _parser = parser;
    }

    /**
     * Allow clients to define a set of known tokens for the
     * template verification. Tokens are expected to be qualified
     * in braces. E.g. {url:path}
     * <p/>
     * The template verification will ensure the known tokens in the
     * URL template conforms to a valid format.
     *
     * @param knownTokens The set of known tokens for a valid template.
     */
    public void setKnownTokens( Collection knownTokens )
    {
        _knownTokens = knownTokens;
    }

    /**
     * Allow clients to define a set of required tokens for the
     * template verification. Tokens are expected to be qualified
     * in braces. E.g. {url:path}
     * <p/>
     * The template verification will ensure the URL template conforms to
     * a valid format for known tokens and contains the required tokens.
     *
     * @param requiredTokens The set of required tokens in a valid template.
     */
    public void setRequiredTokens( Collection requiredTokens )
    {
        _requiredTokens = requiredTokens;
    }

    /**
     * Parses the URL template config file, WEB-INF/url-template-config.xml,
     * and returns the templates.
     *
     * @return The URL templates found in the config file.
     */
    public URLTemplates getTemplates( ServletContext servletContext )
    {
        URLTemplates urlTemplates = new URLTemplates();
        InputStream stream = null;

        try
        {
            UrlTemplateConfig urlTemplateConfig = null;
            stream = servletContext.getResourceAsStream( _configFilePath );
            if ( stream != null )
            {
                urlTemplateConfig = _parser.parse( stream );
                urlTemplates = getTemplatesFromConfig( urlTemplateConfig );
            }
            else
            {
                // No descriptor
                _log.info( "Running without URL template descriptor, " + _configFilePath );
            }
        }
        catch ( XmlException xe )
        {
            // Bad descriptor
            _log.error( "Malformed URL template descriptor in " + _configFilePath, xe );
        }
        catch ( IOException ioe )
        {
            // Bad descriptor
            _log.error( "Problem parsing URL template descriptor in " + _configFilePath, ioe );
        }
        finally
        {
            // Close the stream
            if ( stream != null )
            {
                try
                {
                    stream.close();
                }
                catch ( Exception ignore )
                {
                }
            }
        }

        return urlTemplates;
    }

    /**
     * Loads the templates from a parsed URL template config document.
     *
     * @param urlTemplateConfig the parsed XMLBean document for the URL template config.
     * @return The URL templates found in the config document.
     */
    protected URLTemplates getTemplatesFromConfig( UrlTemplateConfig urlTemplateConfig )
    {
        URLTemplates urlTemplates = new URLTemplates();

        // Load templates
        UrlTemplateDocument.UrlTemplate[] templates = urlTemplateConfig.getUrlTemplateArray();
        for ( int i = 0; i < templates.length; i++ )
        {
            String name = templates[i].getName().trim();
            String value = templates[i].getValue().trim();
            if ( value != null )
            {
                value = value.trim();
                if ( _log.isDebugEnabled() )
                {
                    _log.debug( "[URLTemplate] " + name + " = " + value );
                }
                URLTemplate urlTemplate = new URLTemplate( value, _knownTokens, _requiredTokens );
                urlTemplate.verify();
                urlTemplates.addTemplate( name, urlTemplate );
            }
        }

        // Load template refs
        UrlTemplateRefGroupDocument.UrlTemplateRefGroup[] templateRefGroups = urlTemplateConfig.getUrlTemplateRefGroupArray ();
        for ( int i = 0; i < templateRefGroups.length; i++ )
        {
            HashMap refGroup = new HashMap();
            String refGroupName = templateRefGroups[i].getName().trim();
            UrlTemplateRefDocument.UrlTemplateRef[] templateRefs = templateRefGroups[i].getUrlTemplateRefArray();
            for ( int j = 0; j < templateRefs.length; j++ )
            {
                String key = templateRefs[j].getKey().toString().trim();
                String name = templateRefs[j].getTemplateName().trim();
                if ( _log.isDebugEnabled() )
                {
                    _log.debug( "[" + refGroupName + " URLTemplate] " + key + " = " + name );
                }
                refGroup.put( key, name );
            }
            urlTemplates.addTemplateRefGroup( refGroupName, refGroup );
        }

        return urlTemplates;
    }
}
