package org.apache.beehive.wsm.model.jsr181;

/*
 * Copyright 2005 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Header:$Factory
 */

import java.io.File;
import java.io.FilenameFilter;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import com.sun.mirror.apt.AnnotationProcessorEnvironment;
import com.sun.mirror.apt.Filer;

import java.net.URL;

import org.apache.beehive.wsm.model.BeehiveWsTypeMetadata;

/**
 * Encapsulates all knowledge about where and how object models are persisted.
 */
public class Jsr181ObjectModelStore {
    
    private final static String EXTENSION = ".ser";
    private final static String LOCATOR = ".webservices";

    private static boolean isMarked = false;
    
    private AnnotationProcessorEnvironment env;
    
    /**
     * Constructor.
     * @param env A context that is required to access the filer.
     * TODO needs to be abstracted further.
     */
    public Jsr181ObjectModelStore(AnnotationProcessorEnvironment env) {
        if (null == env) {
            throw new IllegalArgumentException("illegal apt environment: <null>");
        }
        this.env = env;
    }
    
    /**
     * Encapsulates the encoding of the pathname for object models.
     * @param className
     * @return File
     */
    public static File getLocation(String className) {
        return new File(className + EXTENSION);
    }

    /**
     * @param clazz
     * @throws IOException
     * @throws ClassNotFoundException
     */
    public static BeehiveWsTypeMetadata load(Class clazz)
            throws IOException, ClassNotFoundException
    {
        String resourceName = getLocation(clazz.getName()).toString();
        URL url = clazz.getClassLoader().getResource(resourceName);
        return load(url.openStream());
    }
    
    /**
     * @param is
     * @throws IOException
     * @throws ClassNotFoundException
     */
    public static BeehiveWsTypeMetadata load(InputStream is)
            throws IOException, ClassNotFoundException
    {
        Jsr181TypeMetadataImpl objectModel = null;
        ObjectInputStream ois = null;

        try {
            if (null == is) {
                throw new IOException("cannot load object model without input stream");
            }
            ois = new ObjectInputStream(is);
            objectModel = (Jsr181TypeMetadataImpl) ois.readObject();
        }
        finally {
            if (null != ois) {
                ois.close();
            }
        }

        return objectModel;
    }
    
    /**
     * Persists a given <code>objectModel</code>.
     * @param objectModel The object mode to be persisted.
     * @throws IOException If the object model cannot be persisted.
     */
    public void store(BeehiveWsTypeMetadata objectModel)
            throws IOException
    {
        OutputStream os = env.getFiler().createBinaryFile(Filer.Location.CLASS_TREE, "", getLocation(objectModel.getClassName()));
        ObjectOutputStream oos = null;
        try {
            if (null == os) {
                throw new IOException("cannot persist object model without output stream");
            }
            if (null == objectModel) {
                throw new IOException("cannot persist empty object model ");
            }
            oos = new ObjectOutputStream(os);
            oos.writeObject(objectModel);
            
            if (! isMarked) {
                env.getFiler().createBinaryFile(Filer.Location.CLASS_TREE, "", new java.io.File(LOCATOR));
                isMarked = true;
            }
        }
        finally {
            if (null != oos) {
                oos.flush();
                oos.close();
            }
        }
    }

    /**
     * Tries to find the class with a given <code>className</code>. If
     * <code>className</code> is a simple name, the method tries to find a 
     * matching fully qualified class name. The respective class is then
     * loaded and returned.
     * @param className A class name; may be a fully qualified or a simple class
     *   name.
     * @return A class with a matching name.
     */
    public static Class loadWebServiceClass(String className)
            throws ClassNotFoundException
    {
        Class clazz = null;
        ClassLoader cl = Jsr181ObjectModelStore.class.getClassLoader();

        // try className as fully qualified class name
        try {
            clazz = cl.loadClass(className);
        }
        
        // try className as simple class name
        catch (ClassNotFoundException e) {
            final File wsDirectory = new File(new File(cl.getResource(LOCATOR).getFile()).getParent());
            String fqClassName = getFullyQualifiedClassName(wsDirectory, className);
            // todo: we do this for the TCK only...
            if (null == fqClassName) {
                fqClassName = getFullyQualifiedClassName(wsDirectory, className.substring(0, className.length() - "Service".length()));
            }
            clazz = cl.loadClass(fqClassName);
        }
        return clazz;
    }
    
    /**
     * Implements the algorithm for finding a class given a <code>simpleName
     * </code>: searches the directory used for persisting object models for
     * matching names.
     * TODO currently returns the first match; would an exception be the better
     *   approach?
     * @param wsDirectory The directory to be searched.
     * @param simpleName A simple name for the class.
     * @return The fully qualified class name if it can be found.
     * @throws ClassNotFoundException If a matching fully qualified class name 
     *   cannot be found.
     */
    private static String getFullyQualifiedClassName(final File wsDirectory, final String simpleName)
            throws ClassNotFoundException
    {
        FilenameFilter filenameFilter = new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return (wsDirectory.equals(dir) && (name.endsWith(simpleName + EXTENSION)));
            }
        };
        String fqClassName = null;
        for (File file : wsDirectory.listFiles(filenameFilter)) {
            String filename = file.getName();
            if (filename.endsWith(EXTENSION)) {
                fqClassName = filename.substring(0, filename.length() - EXTENSION.length());
                break;
            }
        }
        return fqClassName;
    }
}
