package org.apache.beehive.wsm.model.jsr181;

/*
 * Copyright 2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Header:$Factory
 */

import org.apache.beehive.wsm.model.*;
import org.apache.beehive.wsm.model.java.JavaMethodInfo;
import org.apache.beehive.wsm.model.java.JavaTypeInfo;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.jws.HandlerChain;
import javax.jws.WebMethod;
import javax.jws.WebService;
import javax.jws.soap.SOAPBinding;
import javax.jws.soap.SOAPMessageHandler;
import javax.jws.soap.SOAPMessageHandlers;
import javax.xml.parsers.DocumentBuilderFactory;
import java.io.File;
import java.lang.annotation.Annotation;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

/**
 *
 */
public class Jsr181TypeMetadataImpl implements BeehiveWsTypeMetadata, java.io.Serializable {

    private static final long serialVersionUID = 1L;

    private String wsName;
    private String wsServiceName;
    private String wsWsdlLocation;
    private String wsTargetNamespace;
    private String wsEndpointInterface;
    private Map<String, BeehiveWsMethodMetadata> methodMap =
            new HashMap<String, BeehiveWsMethodMetadata>();
    private Set<BeehiveWsMethodMetadata> noDuplicateMethods;
    private BeehiveWsSOAPBindingInfo soapBinding;
    private String hcFileName;
    private String hcName;
    private List<BeehiveWsSOAPMessageHandlerInfo> soapHandlers =
            new ArrayList<BeehiveWsSOAPMessageHandlerInfo>();
    private String siValue;
    private String className;
    private File baseLocation;

    /**
     * Default constructor.
     */
    public Jsr181TypeMetadataImpl() {
        // empty
    }

    /**
     * This constructor creates an object that encapsulates all type-level
     * annotations and their fields. Moreover, it enforces all rules specified 
     * in JSR-181.
     * @param jt An object that provides Java-specific context services for
     *   type-level annotations.
     */
    public Jsr181TypeMetadataImpl(JavaTypeInfo jt) {

        super();

        if (null == jt) {
            jt.logError("illegal java type info: <null>");
            return;
        }

        // check required @WebService annotation
        WebService webService = jt.getAnnotation(WebService.class);
        if (null == webService) {
            jt.logError("no @WebService annotation found");
        }

        // get webMethods
        Collection<BeehiveWsMethodMetadata> webMethods =
            new ArrayList<BeehiveWsMethodMetadata>();
        // SIB: try adding only annotated methods
        if (! jt.isInterface()) {
            for (JavaMethodInfo jm : jt.getMethods()) {
                if (null != jm.getAnnotation(WebMethod.class)) {
                    BeehiveWsMethodMetadata wsmm = new Jsr181MethodMetadataImpl(jm);
                    if (null == wsmm) {
                        jt.logError("illegal web method: " + jm.getMethodName());
                    }
                    webMethods.add(wsmm);
                }
            }
        }
        // SEI or SIB with no @WebMethod annotations: add all methods
        if (webMethods.isEmpty()) {
            for (JavaMethodInfo jm : jt.getMethods()) {
                BeehiveWsMethodMetadata wsmm = new Jsr181MethodMetadataImpl(jm);
                webMethods.add(wsmm);
            }
        }        
        
        // check required parameters: annotations, webMethods
        Collection<Annotation> annotations = jt.getAnnotations();
        if ((null == annotations) || (null == webMethods)) {
            jt.logError("no annotations or web methods found");
        }

        // check required parameter: className
        className = jt.getName();
        if ((null == className) || (0 == className.length())) {
            jt.logError("web service class is null");
        }
        baseLocation = jt.getLocation();

        // enforce JSR-181 rules
        if (jt.isFinal() || ! jt.isPublic()) {
            jt.logError("@WebService class must be public but not final");
        }
        if (jt.hasFinalize()) {
            jt.logError("web service class must not declare method: finalize()");
        }
        if (jt.isInterface()) {
            if (null != webService.endpointInterface() && 0 < webService.endpointInterface().trim().length()) {
                jt.logError("@WebService.endpointInterface not allowed on interfaces");
            }
            if (null != webService.serviceName() && 0 < webService.serviceName().trim().length()) {
                jt.logError("@WebService.serviceName not allowed on interfaces");
            }
        }
        else {
            if (jt.isAbstract()) {
                jt.logError("service implementation bean must not be abstract");
            }
            if (! jt.hasDefaultConstructor()) {
                jt.logError("service implementation bean must have default constructor");
            }
        }
        if (null != jt.getAnnotation(SOAPMessageHandlers.class) && null != jt.getAnnotation(HandlerChain.class)) {
            jt.logError("Illegal combination of @SOAPMessageHandlers and @HandlerChain");
        }
        String wsdlLocation = webService.wsdlLocation();
        if (null != wsdlLocation && 0 < wsdlLocation.trim().length()) {
            try {
                findResource(wsdlLocation.trim(), baseLocation).openStream();
            }
            catch (Throwable t) {
                jt.logError("wsdlLocation does not exist: " + wsdlLocation);
            }
        }
        
        // initilize instance
        initFromAnnotation(webService);

        // set optional annotations
        for (Annotation a : annotations) {

            // @HandlerChain
            if (a.annotationType() == javax.jws.HandlerChain.class) {
                initFromAnnotation((javax.jws.HandlerChain) a);
            }

            // @SOAPBinding
            else if (a.annotationType() == javax.jws.soap.SOAPBinding.class) {
                initFromAnnotation((javax.jws.soap.SOAPBinding) a);
            }

            // @SOAPMessageHandlers
            else if (a.annotationType() == javax.jws.soap.SOAPMessageHandlers.class) {
                initFromAnnotation((javax.jws.soap.SOAPMessageHandlers) a);
            }

            // @WebService
            else if (a.annotationType() == javax.jws.WebService.class) {
                // ignore -- already done (see above)
                // todo: why can't we do this here?
            }
            
            // unsupported annotation
            else {
                jt.logError("found unsupported annotation: " + a.annotationType().getName());
            }
        }
        
        // set Jsr181MethodMetadata and back fill JSR-181 default values
        for (BeehiveWsMethodMetadata wsmm : webMethods) {
            // JSR-181 defaults: fill in default for tergetNamespace from @WebService
            String wrTargetNamespace = wsmm.getWrTargetNamespace();
            if ((null == wrTargetNamespace) || (0 == wrTargetNamespace.length())) {
                wsmm.setWrTargetNamespace(getWsTargetNamespace());
            }

            // JSR-181 defaults: fill in default for tergetNamespace from @WebService
            for (BeehiveWsParameterMetadata wspm : wsmm.getParams()) {
                String wpTargetNamespace = wspm.getWpTargetNamespace();
                if ((null == wpTargetNamespace) || (0 == wpTargetNamespace.length())) {
                    wspm.setWpTargetNamespace(getWsTargetNamespace());
                }
            }

            // add WebMethod to TYPE metadata
            try {
                addMethod(wsmm);
            }
            catch (ValidationException e) {
                jt.logError("cannot add web method to metadata: " + wsmm.getJavaMethodName());
            }
        }
    }

    protected void initFromAnnotation(WebService annotation) {

        if (null == annotation) {
            return;
        }
        
        // initialize fields with values from annotations
        wsEndpointInterface = annotation.endpointInterface().trim();
        
        setWsName(annotation.name());
        setWsTargetNamespace(annotation.targetNamespace());
        setWsWsdlLocation(annotation.wsdlLocation());
        setWsServiceName(annotation.serviceName());
        setWsEndpointInterface(annotation.endpointInterface());

        // set default values
        
        // name
        String name = className;
        if (-1 < name.indexOf('.')) {
            int index = name.lastIndexOf('.');
            name = name.substring(index + 1);
        }
        if (0 == getWsName().length()) {
            setWsName(name);
        }

        // serviceName
        String serviceName = className;
        if (-1 < serviceName.indexOf('.')) {
            int index = serviceName.lastIndexOf('.');
            serviceName = serviceName.substring(index + 1);
        }
        if (0 == getWsServiceName().length()) {
            setWsServiceName(serviceName + "Service");
        }

        // targetNamespace
        if (0 == getWsTargetNamespace().length()) {
            setWsTargetNamespace(getTargetNamespace(className));
        }
    }

    private void initFromAnnotation(SOAPBinding annotation) {

        if (null == annotation) {
            return;
        }
        
        // set soap binding from annotation
        setSoapBinding(new SOAPBindingInfo(annotation));

        // validate soap binding
        BeehiveWsSOAPBindingInfo sbi = getSoapBinding();
        if (null == sbi) {
            throw new IllegalArgumentException("illegal SOAPBinding annotation on WebService class:\n\t" + sbi);
        }

        // no defaults to handle
    }

    private void initFromAnnotation(HandlerChain annotation) {

        if (null == annotation) {
            return;
        }
        
        if (! getSoapHandlers().isEmpty()) {
            throw new IllegalArgumentException("@SOAPMessageHandlers and @HandlerChain are mutually exclusive");
        }
        
        String hcPath = annotation.file();
        String hcName = annotation.name();

        setHcName(hcName);
        setHcFileName(hcPath);

        try {
            configureHandlerChain(hcPath, hcName);
        }
        catch (Throwable t) {
            throw new RuntimeException(t.getMessage());
        }
    }

    private void initFromAnnotation(SOAPMessageHandlers annotation) {
        
        if (null == annotation) {
            return;
        }
        
        if (null != getHcFileName()) {
            throw new IllegalArgumentException("@SOAPMessageHandlers and @HandlerChain are mutually exclusive");
        }                
                
        for (SOAPMessageHandler soapMessageHandler : annotation.value()) {
            BeehiveWsSOAPMessageHandlerInfo smhi = new SOAPMessageHandlerInfo(soapMessageHandler);
            // todo: validate smhi.getClassName()
            addSOAPHandler(smhi);
        }
    }


    private void configureHandlerChain(String handlerChainConfigPath, String handlerChainName)
        throws Exception
    {
        // get URL for handler-chain config
        URL handlerChainConfigURL = findResource(handlerChainConfigPath, baseLocation);
        if (null == handlerChainConfigURL) {
            throw new Exception("@HandlerChain: cannot resolve relative URL: " + handlerChainConfigPath + "(base directory: " + baseLocation + ")");
        }
        
        // parse handler-chain config and add handler-chain to object model
        initHandlersFromChainConfig(handlerChainConfigURL, handlerChainName);
    }

    /**
     * Finds a resource given a resource path. Tries absolute path (http://...
     * or file:/...) first, then relative to basePath. This method does not use 
     * class loaders since it needs to work with source files as well as with
     * (binary) class files.
     * @param resourcePath the path of the resource, either absolute or relative
     *   (to basePath).
     * @param basePath only required when resourcePath is relative.
     * @return URL of resource.
     */
    private URL findResource(String resourcePath, File basePath) {

        URL resourceURL = null;

        // validate path
        if (null == resourcePath || 0 == resourcePath.length()) {
            return null;
        }
        
        // handle absolute URL
        if (resourcePath.startsWith("http://") || resourcePath.startsWith("file:/")) {
            try {
                resourceURL = new URL(resourcePath);
                return resourceURL;
            }
            catch (MalformedURLException e) {
                return null;
            }
        }
        
        // handle relative URL (we cannot use class loaders here)
        if ((null == basePath) || (! basePath.exists())) {
            return null;
        }
        if (resourcePath.startsWith("/")) {
            String base = basePath.toString();
            String temp = className;
            int lastIdx = className.lastIndexOf('.');
            if (0 >= lastIdx) {
                return null;
            }
            temp = temp.substring(0, lastIdx).replace('.', File.separatorChar);
            if (! base.endsWith(temp)) {
                return null;
            }
            temp = base.substring(0, base.length() - temp.length());
            try {
                resourceURL = new File(temp, resourcePath).toURL();
            }
            catch (MalformedURLException e) {
                return null;
            }
        }
        else {
            try {
                resourceURL = new File(basePath, resourcePath).toURL();
            }
            catch (MalformedURLException e) {
                return null;
            }
        }

        return resourceURL;
    }
    
    /**
     * Made public so specific implementations can configure handlers relevant
     * to their own environment.  Calls to this are ignored if SOAPHandlers are
     * already configured.
     * e.g. in case the config file is located next to the .jws source file in
     * the context root of an axis webapp rather than in the classpath
     */
    public void initHandlersFromChainConfig(URL chainConfigURL, String chainName)
        throws Exception
    {
        // check input parameters
        if (! getSoapHandlers().isEmpty()) {
            throw new Exception("@HandlerChain: annotation doesn't allow for @SOAPMessageHandlers");
        }
        if (null == chainConfigURL || null == chainName) {
            throw new Exception("@HandlerChain: URL for handler-chain config required");
        }
        
        // create and configure factory
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        factory.setIgnoringComments(true);
        
        // todo: this is broken; validate the document!!!
        
        NodeList handlerChainNodes = factory.newDocumentBuilder().parse(chainConfigURL.openStream()).getElementsByTagName("handler-chain");
        for (int j = 0; j < handlerChainNodes.getLength(); j++) {
            Node handlerChainNode = handlerChainNodes.item(j);
            if (handlerChainNode.getNodeType() == Node.ELEMENT_NODE) {
                Element chainElement = (Element) handlerChainNode;
                NodeList chainNames = chainElement.getElementsByTagName("handler-chain-name");
                
                // should only be one chain name so just grab the first one
                // todo: this is broken; validate!!!
                if (chainNames.getLength() > 0 && chainName.equals(chainNames.item(0).getFirstChild().getNodeValue())) {
                    NodeList handlers = chainElement.getElementsByTagName("handler");
                    for (int k = 0; k < handlers.getLength(); k++) {
                        processHandlerNode(handlers.item(k));
                    }
                }
            }
        }
    }

    /**
     * @param handlerNode
     * todo: hack: use "endsWith()" to ignore namespace
     */
    private void processHandlerNode(Node handlerNode)
        throws Exception
    {
        if (handlerNode.hasChildNodes()) {
            String handlerName = null;
            String handlerClassName = null;
            Map<String, String> initParams = new HashMap<String, String>();
            Collection<String> roles = new ArrayList<String>();
            Collection<String> headers = new ArrayList<String>();
            Node currentChild = handlerNode.getFirstChild();
            while (currentChild != null) {
                String nodeName = currentChild.getNodeName();
                if (nodeName.endsWith("handler-name")) {
                    handlerName = currentChild.getFirstChild().getNodeValue();
                }
                else if (nodeName.endsWith("handler-class")) {
                    handlerClassName = currentChild.getFirstChild().getNodeValue();
                }
                else if (nodeName.endsWith("soap-role")) {
                    roles.add(currentChild.getFirstChild().getNodeValue());
                }
                else if (nodeName.endsWith("soap-header")) {
                    headers.add(currentChild.getFirstChild().getNodeValue());
                }
                else if (nodeName.endsWith("init-param")) {
                    nodeName = currentChild.getFirstChild().getNodeName();
                    if (nodeName.endsWith("param-name")) {
                        initParams.put(
                            currentChild.getFirstChild().getNodeValue(),
                            currentChild.getLastChild().getNodeValue());
                    }
                    else if (nodeName.endsWith("param-value")) {
                        initParams.put(
                            currentChild.getLastChild().getNodeValue(),
                            currentChild.getFirstChild().getNodeValue()
                        );
                    }
                }
                currentChild = currentChild.getNextSibling();
            }

            addSOAPHandler(new SOAPMessageHandlerInfo(
                    handlerClassName,
                    handlerName,
                    initParams,
                    roles,
                    headers));
        }
    }
    
    /**
     *
     */
    public void validate() {
        // empty
    }

    /**
     * @return Returns the hcFileName.
     */
    public String getHcFileName() {
        return hcFileName;
    }

    /**
     * @param hcFileName The hcFileName to set.
     */
    public void setHcFileName(String hcFileName) {
        this.hcFileName = hcFileName;
    }

    /**
     * @return Returns the hcName.
     */
    public String getHcName() {
        return hcName;
    }

    /**
     * @param hcName The hcName to set.
     */
    public void setHcName(String hcName) {
        this.hcName = hcName;
    }


    /**
     * @return Returns the siRunAs.
     */
    public String getSiValue() {
        return siValue;
    }

    /**
     * @param siValue The siRunAs to set.
     */
    public void setSiValue(String siValue) {
        this.siValue = siValue;
    }

    /**
     * @return Returns the soapBinding.
     */
    public BeehiveWsSOAPBindingInfo getSoapBinding() {
        if (soapBinding == null) {
            soapBinding = new SOAPBindingInfo();
        }
        return soapBinding;
    }

    /**
     * @param soapBinding The soapBinding to set.
     */
    public void setSoapBinding(BeehiveWsSOAPBindingInfo soapBinding) {
        this.soapBinding = soapBinding;
    }

    /**
     * @return Returns the wsName.
     */
    public String getWsName() {
        return wsName;
    }

    /**
     * @param wsName The wsName to set.
     */
    public void setWsName(String wsName) {
        this.wsName = wsName;
    }

    /**
     * @return Returns the wsServiceName.
     */
    public String getWsServiceName() {
        return wsServiceName;
    }

    /**
     * @param wsServiceName The wsServiceName to set.
     */
    public void setWsServiceName(String wsServiceName) {
        this.wsServiceName = wsServiceName;
    }

    /**
     * @return Returns the wsTargetNamespace.
     */
    public String getWsTargetNamespace() {
        return wsTargetNamespace;
    }

    /**
     * @param wsTargetNamespace The wsTargetNamespace to set.
     */
    public void setWsTargetNamespace(String wsTargetNamespace) {
        this.wsTargetNamespace = wsTargetNamespace;
    }

    
    public String[] getTargetNamespaceParts() {
        // strip the "http://" from the targetnamespace
        String namespace=getWsTargetNamespace().substring(7, getWsTargetNamespace().length());
         String[] beforeTranspose = namespace.split("[\\./]");
        String[] res = new String[beforeTranspose.length];
        for(int i=0; i<res.length; i++) 
            res[i] = beforeTranspose[res.length - i - 1];
        return res;
    }
    
    /**
     * @return Returns the wsEndpointInterface.
     */
    public String getWsEndpointInterface() {
        return wsEndpointInterface;
    }

    /**
     * @param wsEndpointInterface The wsEndpointInterface to set.
     */
    public void setWsEndpointInterface(String wsEndpointInterface) {
        this.wsEndpointInterface = wsEndpointInterface;
    }

    /**
     * @return Returns the wsWsdlLocation.
     */
    public String getWsWsdlLocation() {
        return wsWsdlLocation;
    }

    /**
     * @param wsWsdlLocation The wsWsdlLocation to set.
     */
    public void setWsWsdlLocation(String wsWsdlLocation) {
        this.wsWsdlLocation = wsWsdlLocation;
    }

    /**
     * @return Returns the methods.
     */
    public Collection<BeehiveWsMethodMetadata> getMethods() {
        if (noDuplicateMethods == null) {
            noDuplicateMethods = Collections
                .unmodifiableSet(new HashSet<BeehiveWsMethodMetadata>
                                 (methodMap.values()));
        }
        return noDuplicateMethods;
    }

    public BeehiveWsMethodMetadata getMethod(String methodName,
                                          Class... paramTypes) {
        String sig = createCompleteMethodSignature(methodName, paramTypes);
        
        BeehiveWsMethodMetadata meta = methodMap.get(sig);
        if (meta == null) {
            /* 
             * NOTE jcolwell@bea.com 2004-Nov-29 -- 
             * this fallback may be a bad thing but it is nice to have if the
             * WSDLProcessor cannot map the Java types properly when used on the
             * client side.
             */
            meta = methodMap.get(methodName + '('+ paramTypes.length + ')');
        }
        return meta;
    }

    public void addMethod(BeehiveWsMethodMetadata method)
        throws ValidationException {

        // NOTE jcolwell@bea.com 2004-Nov-29 -- clear the Set that prevents
        // duplicates being returned from getMethods().
        noDuplicateMethods = null;
        List<? extends BeehiveWsParameterMetadata> params = method.getParams();
        Class[] paramTypes = new Class[params.size()];
        int j = 0;
        for (BeehiveWsParameterMetadata param : params) {
            Class javaType = param.getJavaType();
            if (null == javaType) {
                throw new ValidationException("<null> is not a valid parameter class");
            }
            paramTypes[j++] = javaType;
        }
        String opName = method.getWmOperationName();
        String sig = createCompleteMethodSignature(opName, paramTypes);
        
        if (methodMap.containsKey(sig)) {
            throw new ValidationException(sig + " is already present in this Web Service and duplicate methods are not permitted");
        }
        else if (SOAPBinding.Style.DOCUMENT.equals(getSoapBinding().getStyle()) && methodMap.containsKey(opName)) {
            throw new ValidationException("document-style does not allow duplicate methods: " + opName);
        }
        else {
            methodMap.put(sig, method);
            methodMap.put(opName + '(' + paramTypes.length + ')', method);
        }
    }

    public List<? extends BeehiveWsSOAPMessageHandlerInfo> getSoapHandlers() {
        return Collections.unmodifiableList(soapHandlers);
    }

    public void addSOAPHandler(BeehiveWsSOAPMessageHandlerInfo soapHandler) {
        soapHandlers.add(soapHandler);
    }

    public String getClassName() {
        return className;
    }

    public void setClassName(String className) {
        this.className = className;
    }

    private String getTargetNamespace(String fqClassName) {
        if ((null == fqClassName) || (0 == fqClassName.length())) {
            return fqClassName;
        }
        String[] tokens = fqClassName.split("\\.");
        String targetNamespace = tokens[0];
        for (int i = 1; i <= tokens.length - 2; i++) {
            targetNamespace = tokens[i] + "." + targetNamespace;
        }
        return "http://" + targetNamespace;
    }

    // todo: this needs to go, as it should really be in Jsr181MethodMetadataImpl!!!
    private String createCompleteMethodSignature(String name,
                                                 Class... paramTypes) {

        StringBuilder sb = new StringBuilder(name);
        sb.append('(');
        boolean firstParam = true;
        if (paramTypes != null) {
            for (Class type : paramTypes) {
                if (firstParam) {
                    firstParam = false;
                }
                else {
                    sb.append(',');
                }
                sb.append(type.getName());
            }
        }
        sb.append(')');
        return sb.toString();
    }

    /**
     * @param jt SIB type info.
     * @throws Exception
     */
    public void merge(JavaTypeInfo jt) throws Exception {
        
        // check if this has endpointInterface --> error
        String myEndpointInterface = getWsEndpointInterface();
        if (null != myEndpointInterface && 0 < myEndpointInterface.length())
            throw new ValidationException("service endpoint interface can't reference another service endpoint interface: " + myEndpointInterface);
        
        // check if objectModel doesn't reference this endpointInterface --> error
        BeehiveWsTypeMetadata sibObjectModel = new Jsr181TypeMetadataImpl(jt);
        String omEndpointInterface = sibObjectModel.getWsEndpointInterface();
        if (null == omEndpointInterface || 0 >= omEndpointInterface.length() || ! omEndpointInterface.equals(getClassName()))
            throw new Exception("Internal error: object model for " + sibObjectModel.getClassName() + " does not reference endpoint interface "+ omEndpointInterface);

        // check if objectModel.wsWsdlLocation != null --> error
        if (null != sibObjectModel.getWsWsdlLocation() && 0 < sibObjectModel.getWsWsdlLocation().length()) {
            throw new Exception("implementation bean must not reference a service endpoint interface and a WSDL location at the same time");
        }
        
        setWsServiceName(sibObjectModel.getWsServiceName());
        setWsEndpointInterface(sibObjectModel.getWsEndpointInterface());
        
        setClassName(sibObjectModel.getClassName());
        
        // check if jt implements the contract
        validateContract(jt);
    }

    /**
     * Checks if "jt" implements "this".
     * @param jt SIB type info.
     * @throws ValidationException
     */
    protected void validateContract(JavaTypeInfo jt) throws ValidationException {
        
        // get all implemented methods
        Collection<BeehiveWsMethodMetadata> implementedMethods = new HashSet<BeehiveWsMethodMetadata>();
        for (JavaMethodInfo jm : jt.getMethods()) {
            try {
                implementedMethods.add(new Jsr181MethodMetadataImpl(jm));
            }
            catch (Throwable t) {
                throw new ValidationException("invalid method declaration found: " + jt.getName() + ": " + jm.getMethodName());
            }
        }
        
        // check if we implement all required methods
        for (BeehiveWsMethodMetadata declaredMethod : getMethods()) {
            boolean implementationFound = false;
            for (BeehiveWsMethodMetadata implementedMethod : implementedMethods) {
                if (implementedMethod.equals(declaredMethod)) {
                    implementationFound = true;
                    break;
                }
            }
            if (! implementationFound) {
                throw new ValidationException("method not implemented by " + jt.getName() + ": " + declaredMethod);
            }
        }
    }
}
