/*
 * 
 * Copyright 2001-2004 The Apache Software Foundation.
 * 
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * 
 */
package org.apache.beehive.wsm.registration;

import java.io.File;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.GenericArrayType;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.rmi.Remote;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.*;

import javax.jws.soap.SOAPBinding;
import javax.xml.namespace.QName;
import javax.xml.rpc.encoding.TypeMapping;
import javax.xml.rpc.holders.Holder;

import org.apache.beehive.wsm.databinding.BindingLookupService;
import org.apache.beehive.wsm.databinding.GenericHolder;

public abstract class TypeRegistrar {
	static Logger logger = Logger.getLogger(TypeRegistrar.class);

	protected TypeMapping mTypeMapping;

	protected BindingLookupService lookupService;

	/**
	 * @param typeMapping
	 * @param lookupService
	 */
	public TypeRegistrar(TypeMapping typeMapping,
			BindingLookupService lookupService) {
		super();
		mTypeMapping = typeMapping;
		this.lookupService = lookupService;
	}

	/**
	 * @param cls
	 * @return QName
	 */
	protected abstract QName getRegisteredQName(Class cls);

	/**
	 * @param cls
	 * @param q
	 */
	abstract protected void registerClassAsWithDefaultSearialization(Class cls,
			QName q, SOAPBinding.Style style,
			SOAPBinding.Use use);

	/**
	 * @param cls
	 * @param q
	 */
	abstract protected void registerClassAsXMLBeans(Class cls, QName q);

	/**
	 * @param cls
	 * @param q
	 * @throws ClassNotFoundException
	 * @throws NoSuchMethodException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	abstract protected void registerClassAsDataHandler(Class cls, QName q)
			throws ClassNotFoundException, NoSuchMethodException,
			InstantiationException, IllegalAccessException,
			InvocationTargetException;

	/**
	 * @param cls
	 * @param q
	 */
	abstract protected void registerClassAsSoapEncodedArray(Class cls, QName q);

	/**
	 * @param cls
	 * @param q
	 * @return boolean
	 */
	abstract protected boolean classIsRegistered(Class cls, QName q);

	/**
	 * @return QName
	 */
	abstract public QName getVoidType();

	abstract protected boolean isBuiltInType(Class cls);

	abstract protected QName getBuiltInTypeQname(Class cls);

	public QName registerType(Class cls, SOAPBinding.Style style,
			SOAPBinding.Use use) {
		QName q;
		if (SOAPBinding.Use.ENCODED == use && cls.isArray()) {
			q = lookupService.class2qname(cls.getComponentType());
		}
		else {
			q = lookupService.class2qname(cls);
		}
		return registerType(cls, q, style, use);
	}

	public QName registerType(Class cls, QName q, SOAPBinding.Style style,
			SOAPBinding.Use use) {
		try {
			logger.debug("Register class: " + cls.getCanonicalName()
					+ " qName: " + q);
			if (Void.TYPE.equals(cls))
				return null;

			if (isBuiltInType(cls))
				return getBuiltInTypeQname(cls);
			if (q == null)
				throw new RuntimeException(
						"Invalid registeration requestion qname is null");

			if (cls.isArray()) {
				if (SOAPBinding.Use.ENCODED == use) {

					// The Qname modification was needed to run the GoogleThe qname of the array must be different from the qname of the element.
					// I don't think the qname really matters, as the array qname wont be in the message
					// The qname of the array is available in the WSDL, but at this point we don't
					// have the information, I have added "Array" at the end of the qname, it can
					// be anything else and the code seems to work (from the client side).
					// if this needs to be the real name of the array, somehow WSDL information
					// must be made available to this method.
					QName arrayQname = new QName(q.getNamespaceURI(), q.getLocalPart()+"Array"); 

					if (!classIsRegistered(cls, arrayQname))
						registerClassAsSoapEncodedArray(cls, arrayQname);
				}
				// incase of literal, or encoded register the class also
				q = registerType(cls.getComponentType(), q, style, use);

			} else if (!classIsRegistered(cls, q)) {
				if (org.apache.xmlbeans.XmlObject.class.isAssignableFrom(cls)) {
					registerClassAsXMLBeans(cls, q);
				}
				/*
				 * NOTE jcolwell@bea.com 2004-Oct-11 -- these datahandler using
				 * classes are generally already registered but just in case...
				 */
				else if (isActivationEnabled()
						&& (java.awt.Image.class.isAssignableFrom(cls)
								|| getMultipartClass().isAssignableFrom(cls) || getDataHandlerClass()
								.isAssignableFrom(cls))) {
					try {
						registerClassAsDataHandler(cls, q);
					} catch (Exception e) {
						/*
						 * FIXME jcolwell@bea.com 2004-Oct-08 -- log this
						 * properly
						 */
						e.printStackTrace();
					}
				} else if (!Remote.class.isAssignableFrom(cls)
				/*
				 * NOTE jcolwell@bea.com 2004-Oct-11 -- java.rmi.Remote is
				 * prohibited by the jax-rpc spec.
				 */

				/*
				 * NOTE jcolwell@bea.com 2004-Oct-11 -- restricting against File
				 * since, it doesn't make sense to serialize as a bean. That and
				 * it causes an infinite loop as it keeps returning itself from
				 * the getAbsoluteFile and getCanonicalFile calls
				 */
				&& !File.class.isAssignableFrom(cls)) {
					registerClassAsWithDefaultSearialization(cls, q, style, use);
				} else {
					throw new RuntimeException("failed to register "
							+ cls.getName()
							+ " as a valid web service datatype,"
							+ " consider using a custom type mapping");
				}
			}
			// }
			// }
			logger.info("Registered class: " + cls.getCanonicalName()
					+ " qName: " + q);
			return q;
		} catch (RuntimeException e) {
			logger.error("Failed to register class: " + cls.getCanonicalName()
					+ " type: " + q.getNamespaceURI() + ":" + q.getLocalPart());

			e.printStackTrace();
			throw e;
		}
	} // public Class q2Class(QName qType) { //

	private boolean isActivationEnabled() {
		return null != getDataHandlerClass() && null != getMultipartClass();
	}

	private Class getDataHandlerClass() {
		try {
			return getClass().getClassLoader().loadClass(
					"javax.activation.DataHandler");
		} catch (Exception e) {
		}
		return null;
	}

	private Class getMultipartClass() {
		try {
			return getClass().getClassLoader().loadClass(
					"javax.mail.internet.MimeMultipart");
		} catch (Exception e) {
		}
		return null;
	}

	/**
	 * @param t
	 * @throws Exception
	 */
	static public Class getHoldersValueClass(Type t) {
		Class res = null;

		if (t instanceof ParameterizedType) {
			ParameterizedType pt = (ParameterizedType) t;
			Type raw = pt.getRawType();
			if (GenericHolder.class.isAssignableFrom((Class) raw)) {
				Type[] typeArgs = pt.getActualTypeArguments();
				if (typeArgs[0] instanceof GenericArrayType) {
					Class arrayElementType = (Class) ((GenericArrayType) typeArgs[0])
							.getGenericComponentType();
					res = Array.newInstance(arrayElementType, 0).getClass();
				} else {
					res = (Class) typeArgs[0];
				}
			} else {
				throw new RuntimeException(
						"Invalid parameterized type for holder: " + t);
			}

		} else if (t instanceof Class) {
			if (Holder.class.isAssignableFrom((Class) t)) {
				// if type is not a holder type, there is no point to dig any
				// deeper
				Field[] publicFields = ((Class) t).getFields();
				for (int i = 0; i < publicFields.length; i++) {
					if (0 == "value".compareTo(publicFields[i].getName())) {
						res = publicFields[i].getType();
						break;
					}
				}
			} else {
				throw new RuntimeException("Invalid class.  Type: " + t
						+ " is not a holder.");
			}
		}
		// other wise just return the type
		return res;
	}
}
